/*
	   Copyright (C) 2012 Gilles Landais (CDS)

    Author: Gilles Landais, Strasbourg astronomical Data Center (CDS)
    Email: gilles.landais@unistra.fr

    This file is part of H3C.

    H3C is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    H3C is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with H3C; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


/** @file h3c_poly.c
 *  @brief deal with polynome
 *  @author 2004-2008 Sergey Koposov (Q3C), Gilles Landais
 *
 *  @note a part of this file contains function coming from the Q3C library
 *
	   Copyright (C) 2004-2008 Sergey Koposov

    Author: Sergey Koposov, Sternberg Astronomical Institute
    Email: math@sai.msu.ru
    http://lnfm1.sai.msu.ru/~math

    This file is part of Q3C.

    Q3C is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Q3C is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Q3C; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


#include "common.h"

#if 0
void h3c_init_poly(h3c_poly *qp, int n)
{
	qp->ra = malloc(n * sizeof(h3c_coord_t));
	qp->dec = malloc(n * sizeof(h3c_coord_t));

	qp->ax = malloc(n * sizeof(h3c_coord_t));
	qp->ay = malloc(n * sizeof(h3c_coord_t));

	qp->x = malloc(n * sizeof(h3c_coord_t));
	qp->y = malloc(n * sizeof(h3c_coord_t));
	qp->n = n;
}
#endif
static void h3c_prepare_poly(h3c_poly *qp)
{
	int n = qp->n - 1 ;
	int i;
	h3c_coord_t *ax = qp->ax;
	h3c_coord_t *ay = qp->ay;
	h3c_coord_t *x = qp->x;
	h3c_coord_t *y = qp->y;

	for(i = 0 ; i < n; i++)
	{
		ax[i] = x[i + 1] - x[i];
		ay[i] = y[i + 1] - y[i];
	}
	ax[i] = x[0] - x[i];
	ay[i] = y[0] - y[i];
}

/* Cloned version of ang2ipix for outputting also the x,y on the cube face
 * Coordinates on the cube face are x[-0.5,0.5] y[-0.5,0.5] */
static void ang2ipix_xy (/*struct q3c_prm *hprm,*/ h3c_coord_t ra, h3c_coord_t dec,
					char *out_face_num, /*q3c_ipix_t *ipix, */h3c_coord_t *x_out,
					h3c_coord_t *y_out)
					/* ra in degrees, dec in degrees */
					/* strictly 0<=ra<360 and -90<=dec<=90 */
{
	h3c_coord_t x0 = 0,y0 = 0;
	/*q3c_ipix_t nside = hprm->nside, *xbits = hprm->xbits,
				*ybits = hprm->ybits, xi, yi, i1;*/
	char face_num;
	if (dec >= 90)
	/* Poles */
	{
		face_num = 0;
		x0 = H3C_HALF;
		y0 = H3C_HALF;
		*x_out = 0;
		*y_out = 0;
		goto END1;
	}
	else if (dec <= -90)
	{
		face_num = 5;
		x0 = H3C_HALF;
		y0 = H3C_HALF;
		*x_out = 0;
		*y_out = 0;
		goto END1;
	}
	
	face_num = fmod ((ra + 45) / 90, 4);
	/*for equatorial pixels we'll have face_num from 1 to 4 */
	x0 = tan (H3C_DEGRA * (ra - 90 * (h3c_coord_t)face_num));
	y0 = tan (dec * H3C_DEGRA) /
		cos (H3C_DEGRA * (ra - 90 * (h3c_coord_t)face_num));
	face_num++;
	
	if (y0 > 1)
	{
 		face_num = 0;
		x0 = sin (H3C_DEGRA * ra) / tan (H3C_DEGRA * dec);
		y0 = -1*cos (H3C_DEGRA * ra) / tan (H3C_DEGRA * dec);
	}
	else if (y0 < -1)
	{
		face_num = 5;
		x0 = -1*sin (H3C_DEGRA * ra) / tan (H3C_DEGRA * dec);
		y0 = -1*cos (H3C_DEGRA * ra) / tan (H3C_DEGRA * dec);
	}
	
	*x_out = x0 / 2;
	*y_out = y0 / 2;
	x0 = (x0 + 1) / 2;
	y0 = (y0 + 1) / 2;
	
	END1:
	
	/* Now I produce the final pixel value by converting x and y values to bitfields
	 * and combining them by interleaving, using the predefined arrays xbits and ybits
	 */
#if 0	
	xi = (q3c_ipix_t)(x0 * nside);
	yi = (q3c_ipix_t)(y0 * nside);
	
	/* This two strings are written to handle the case of edges of base square */
	if (xi == nside)
	{
		xi--;
	}
	if (yi == nside)
	{
		yi--;
	}
	
	i1 = 1 << (Q3C_INTERLEAVED_NBITS);
	
	
#ifdef Q3C_INT4
	*ipix = ((q3c_ipix_t)face_num) * nside * nside + xbits[xi % i1] +
		ybits[yi % i1];
	/*4byte computation*/
#endif /* Q3C_INT4 */
#ifdef Q3C_INT8
	*ipix = ((q3c_ipix_t)face_num) * nside * nside + xbits[xi % i1] +
		ybits[yi % i1] + (xbits[(xi >> Q3C_INTERLEAVED_NBITS) % i1] +
		ybits[(yi >> Q3C_INTERLEAVED_NBITS) % i1]) * i1 * i1;
	/*8byte computation*/
#endif /* Q3C_INT8 */
#endif	
	*out_face_num = face_num;
}

static int h3c_check_point_in_poly(h3c_poly *qp, h3c_coord_t x0,
                            h3c_coord_t y0)
/* Implementation of the crossing algorithm */
{
	int i, n = qp->n;
	h3c_coord_t *y = qp->y;
	h3c_coord_t *x = qp->x;
	h3c_coord_t *ax = qp->ax;
	h3c_coord_t *ay = qp->ay;
	int result = !H3C_DISJUNCT;
	for(i=0;i<n;i++)
	{
		if (((y0<=y[i])==(y0>y[(i+1)%n])) &&
				((x0-x[i])<(y0-y[i])*ax[i]/ay[i]))
		{
			result=!result;
		}
	}
	return !result;
}

static void h3c_get_minmax_poly(h3c_poly *qp, h3c_coord_t *xmin,
                         h3c_coord_t *xmax, h3c_coord_t *ymin,
                         h3c_coord_t *ymax)
{
	int i;
	const int n = qp->n;
	h3c_coord_t *x = qp->x, *y = qp->y, t;
	h3c_coord_t xmi, xma, ymi, yma;

	xmi = x[0];
	xma = x[0];
	ymi = y[0];
	yma = y[0];

	for(i = 1; i < n; i++)
	{
		t = x[i];
		if (t > xma)
		{
			xma = t;
		}
		else if (t < xmi)
		{
			xmi = t;
		}

		t = y[i];

		if (t > yma)
		{
			yma = t;
		}
		else if (t < ymi)
		{
			ymi = t;
		}
	}

	*xmin = xmi;
	*xmax = xma;
	*ymin = ymi;
	*ymax = yma;
}


static char h3c_get_facenum(h3c_coord_t ra, h3c_coord_t dec)
	/* ra in degrees, dec in degrees */
	/* strictly 0<=ra<360 and -90<=dec<=90 */
{
	h3c_coord_t y0 = 0;
	char face_num;
	
	if (dec >= 90)
	/* Poles */
	{
		return 0;
	}
	else if (dec <= -90)
	{
		return 5;
	}
	
	face_num = fmod ((ra + 45) / 90, 4);
	/*for equatorial pixels we'll have face_num from 1 to 4 */
	
	y0 = tan(dec * H3C_DEGRA) /
			cos(H3C_DEGRA * (ra - 90 * (h3c_coord_t)face_num));
	
	face_num++;
	
	if (y0 > 1)
	{
		return 0;
	}
	else if (y0 < -1)
	{
		return 5;
	}
	else
	{
		return face_num;
	}
}
static char h3c_xy2facenum(h3c_coord_t x, h3c_coord_t y, char face_num0)
{
	/* The input x, y should be >=-1  and <=1 */
	
	
	h3c_coord_t ra = 0, dec = 0;
	/* I do the initialization since gcc warn about probable not
	 * initialization of ra and dec
	 */
	
	/* This code have been cutted out from ipix2ang BEGIN */
	if ((face_num0 >= 1) && (face_num0 <= 4))
	{
		ra = atan(x);
		dec = H3C_RADEG * atan(y * cos(ra));
		ra = ra * H3C_RADEG + ((h3c_coord_t)face_num0 - 1) * 90;
		if (ra < 0)
		{
			ra += (h3c_coord_t)360;
		}
	}
	else
	{
		if (face_num0 == 0)
		{
			ra = H3C_RADEG * atan2(x, -y);
			dec = H3C_RADEG * atan(1 / sqrt(x * x + y * y));
			if (ra < 0)
			{
				ra += (h3c_coord_t)360;
			}
		}
		if (face_num0 == 5)
		{
			ra = H3C_RADEG * atan2(x, y);
			dec = -H3C_RADEG * atan(1 / sqrt(x * x + y * y));
			if (ra < 0)
			{
				ra += (h3c_coord_t)360;
			}

		}
	}
	/* This code have been cutted out from ipix2ang END */
	
	return h3c_get_facenum(ra,dec);
}

static char h3c_get_facenum_poly(h3c_poly *qp)
{
	return h3c_get_facenum(qp->ra[0], qp->dec[0]);
}

static void h3c_project_poly(h3c_poly *qp, char face_num)
{
	h3c_coord_t ra1, dec1, tmp0;
	h3c_coord_t *ra = qp->ra, *dec = qp->dec;
	h3c_coord_t *x = qp->x, *y = qp->y, x0, y0;

	int i, n = qp->n;
	if ((face_num > 0) && (face_num < 5))
	{
		face_num--; /* Just computation trick */
		for (i = 0; i < n; i++)
		{
			ra1 = H3C_DEGRA * (ra[i] - 90 * (h3c_coord_t)face_num);
			dec1 = H3C_DEGRA * dec[i];
			x[i] = (tan(ra1)) / 2;
			y[i] = (tan(dec1) / cos(ra1)) / 2;
		}
		/* Now x[i] and y[i] are coordinates on cube face [-0.5:0.5]x[-0.5:0.5] */
	}
	else if (face_num == 0)
	{
		for (i = 0; i < n; i++)
		{
			ra1 = H3C_DEGRA * ra[i];
			dec1 = H3C_DEGRA * dec[i];

			tmp0 = 1 / tan(dec1);
#ifdef __USE_GNU
			sincos(ra1, &x0, &y0);
#else
			x0 = sin(ra1);
			y0 = cos(ra1);
#endif
			x0 *= tmp0;
			y0 *= (-tmp0);
			x[i] = x0 / 2;
			y[i] = y0 / 2;
		}
	}
	else
	{
		for (i = 0; i < n; i++)
		{
			ra1 = H3C_DEGRA * ra[i];
			dec1 = H3C_DEGRA * dec[i];

			tmp0 = 1 / tan(dec1);
#ifdef __USE_GNU
			sincos(ra1, &x0, &y0);
#else
			x0 = sin(ra1);
			y0 = cos(ra1);
#endif
			x0 *= (-tmp0);
			y0 *= (-tmp0);
			x[i] = x0 / 2;
			y[i] = y0 / 2;
		}
	}
}
#if 0


static char h3c_poly_intersection_check(h3c_poly *qp,
	h3c_coord_t xl, h3c_coord_t xr,
	h3c_coord_t yb, h3c_coord_t yt,
	h3c_coord_t cur_size)
{
	int i, n = qp->n;
	h3c_coord_t *ax = qp->ax;
	h3c_coord_t *ay = qp->ay;
	h3c_coord_t *x = qp->x;
	h3c_coord_t *y = qp->y;
	h3c_coord_t txl, txr, tyb, tyt, axi, ayi, xi, yi, tmp, tmp1;
	char ret = 0;
	for( i = 0; i <n ; i++)
	{
		xi = x[i];
		yi = y[i];
		axi = ax[i];
		ayi = ay[i];
		txl = xl - xi;
		txr = xr - xi;
		tyb = yb - yi;
		tyt = yt - yi;

		tmp = tyb / ayi;
		tmp1 = axi * tmp - txl;
		if ((tmp >= 0) && (tmp <= 1) && (tmp1 >= 0) && (tmp1 <= cur_size))
		{
			ret = 1;
			break;
		}

		tmp = tyt / ayi;
		tmp1 = axi * tmp - txl;
		if ((tmp >= 0) && (tmp <= 1) && (tmp1 >= 0) && (tmp1 <= cur_size))
		{
			ret = 1;
			break;
		}

		tmp = txl / axi;
		tmp1 = ayi * tmp - tyb;
		if ((tmp >= 0) && (tmp <= 1) && (tmp1 >= 0) && (tmp1 <= cur_size))
		{
			ret = 1;
			break;
		}

		tmp = txr / axi;
		tmp1 = ayi * tmp - tyb;
		if ((tmp >= 0) && (tmp <= 1) && (tmp1 >= 0) && (tmp1 <= cur_size))
		{
			ret = 1;
			break;
		}
	}

	return ret;
}



int h3c_poly_cover_check(h3c_poly *qp, h3c_coord_t xc_cur,
						h3c_coord_t yc_cur, h3c_coord_t cur_size)
{
	h3c_coord_t xl_cur, xr_cur, yb_cur, yt_cur;
	int val;


	/* Checking the intersection of ellipse and box
	 * The box parameters are set by variables xc_cur, yc_cur and cur_size
	 */
	xl_cur = xc_cur - cur_size / 2; /* left   */
	xr_cur = xc_cur + cur_size / 2; /* right  */
	yb_cur = yc_cur - cur_size / 2; /* bottom */
	yt_cur = yc_cur + cur_size / 2; /* top    */

	/* Undef labels -- the labels when the current computed values dont allow
	 * to make the final decision about the intersection
	 */

	val = h3c_check_point_in_poly(qp, xl_cur, yb_cur);
	if (val != H3C_DISJUNCT)
	{
		goto PARTUNDEF_CHECK01;
	}

	val = h3c_check_point_in_poly(qp, xr_cur, yb_cur);
	if (val != H3C_DISJUNCT)
	{
		return H3C_PARTIAL;
	}

	val = h3c_check_point_in_poly(qp, xr_cur, yt_cur);
	if (val != H3C_DISJUNCT)
	{
		return H3C_PARTIAL;
	}

	val = h3c_check_point_in_poly(qp, xl_cur, yt_cur);

	if (val != H3C_DISJUNCT)
	{
		return H3C_PARTIAL;
	}
	else
	{
		if (h3c_poly_intersection_check(qp, xl_cur, xr_cur, yb_cur, yt_cur, cur_size)||
			((qp->x[0] > xl_cur) && (qp->x[0] < xr_cur) &&
			(qp->y[0] > yb_cur) && (qp->y[0] < yt_cur)))
		{
			return H3C_PARTIAL;
		}
		else
		{
			return H3C_DISJUNCT;
		}
	}


	PARTUNDEF_CHECK01:
	val = h3c_check_point_in_poly(qp, xr_cur, yb_cur);
	if (val == H3C_DISJUNCT)
	{
		return H3C_PARTIAL;
	}

	//PARTUNDEF_CHECK11:
	val = h3c_check_point_in_poly(qp, xr_cur, yt_cur);
	if (val == H3C_DISJUNCT)
	{
		return H3C_PARTIAL;
	}


	//PARTUNDEF_CHECK10:
	val = h3c_check_point_in_poly(qp, xl_cur, yt_cur);
	if (val == H3C_DISJUNCT)
	{
		return H3C_PARTIAL;
	}
	else
	{
		return H3C_COVER;
	}
}
#endif

int h3c_check_sphere_point_in_poly(/*struct h3c_prm *hprm,*/ int n,
                                   h3c_coord_t in_ra[], h3c_coord_t in_dec[],
                                   h3c_coord_t ra0, h3c_coord_t dec0)
{
	h3c_coord_t xmin,xmax,ymin, ymax;
	static char faces[6], multi_flag;
/*	h3c_ipix_t ipix;*/
	h3c_coord_t points[4];
	char face_num, face_num0, cur_face_num;
	static h3c_coord_t x[3][H3C_MAX_N_POLY_VERTEX], y[3][H3C_MAX_N_POLY_VERTEX],
		ax[3][H3C_MAX_N_POLY_VERTEX], ay[3][H3C_MAX_N_POLY_VERTEX], x0, y0;

	int face_count = -1, i;

	h3c_poly qp;

	ang2ipix_xy(/*hprm,*/ ra0, dec0, &cur_face_num,/* &ipix, */&x0, &y0);

	qp.ra = in_ra;
	qp.dec = in_dec;
	qp.n = n;

		face_num = h3c_get_facenum_poly(&qp);
		faces[0] = face_num;

		qp.x = x[0];
		qp.y = y[0];
		qp.ax = ax[0];
		qp.ay = ay[0];

		h3c_project_poly(&qp, face_num);
		h3c_prepare_poly(&qp);

		h3c_get_minmax_poly(&qp, &xmin, &xmax, &ymin, &ymax);

		/* Now in a little bit ugly but fastest way I determine whether the ellipse
		 * intersect other faces or not, and if yes, I setup the array "points" to the
		 * multi_face loop.
		 */

		if (xmin < -H3C_HALF)
		{
			if (ymin < -H3C_HALF)
			{
				points[0] = xmax;
				points[1] = ymin;
				points[2] = xmin;
				points[3] = ymax;
				multi_flag = 2;
				xmin = -H3C_HALF;
				ymin = -H3C_HALF;
			}
			else
			{
				if (ymax > H3C_HALF)
				{
					points[0] = xmax;
					points[1] = ymax;
					points[2] = xmin;
					points[3] = ymin;
					multi_flag = 2;
					xmin = -H3C_HALF;
					ymax = H3C_HALF;
				}
				else
				{
					points[0] = xmin;
					points[1] = (ymin + ymax) / 2;
					multi_flag = 1;
					xmin = -H3C_HALF;
				}
			}
		}
		else
		{
			if (xmax > H3C_HALF)
			{
				if (ymin < -H3C_HALF)
				{
					points[0] = xmin;
					points[1] = ymin;
					points[2] = xmax;
					points[3] = ymax;
					multi_flag = 2;
					xmax = H3C_HALF;
					ymin = -H3C_HALF;
				}
				else
				{
					if (ymax > H3C_HALF)
					{
						points[0] = xmin;
						points[1] = ymax;
						points[2] = xmax;
						points[3] = ymax;
						multi_flag = 2;
						xmax = H3C_HALF;
						ymax = H3C_HALF;
					}
					else
					{
						points[0] = xmax;
						points[1] = (ymin + ymax) / 2;
						multi_flag = 1;
						xmax = H3C_HALF;
					}
				}
			}
			else
			{
				if (ymin < -H3C_HALF)
				{
					points[0] = (xmin + xmax) / 2;
					points[1] = ymin;
					multi_flag = 1;
					ymin = -H3C_HALF;
				}
				else
				{
					if (ymax > H3C_HALF)
					{
						points[0] = (xmin + xmax) / 2;
						points[1] = ymax;
						multi_flag = 2;
						ymax = H3C_HALF;
					}
					else
					{
						multi_flag = 0;
					}
				}
			}
		}

		face_num0 = face_num;

		for(face_count = 0; face_count <= multi_flag; face_count++)
		{
		/* This the beginning of the mega-loop over multiple faces */

			if (face_count > 0)
			/* This "if" works when we pass through the secondary faces */
			{
				face_num = h3c_xy2facenum(2 * points[2 * face_count - 2],
					2 * points[2 * face_count - 1], face_num0);

				faces[face_count] = face_num;

				qp.x = x[face_count];
				qp.y = y[face_count];
				qp.ax = ax[face_count];
				qp.ay = ay[face_count];

				h3c_project_poly(&qp, faces[face_count]);
				h3c_prepare_poly(&qp);
			}
		}

	for (i = 0; i <= multi_flag; i++)
	{
		if (faces[i] == cur_face_num)
		{
			face_count = i;
			break;
		}
	}

	if (i == (multi_flag + 1))
	{
		return 0;
	}

	qp.x = x[face_count];
	qp.y = y[face_count];
	qp.ax = ax[face_count];
	qp.ay = ay[face_count];

	return h3c_check_point_in_poly(&qp, x0, y0);
}

